class ValorantApi(Exception):
    pass


class InvalidOrMissingParameter(ValorantApi):  # 400
    pass


class NotFound(ValorantApi):  # 404
    pass

class AttributeExistsError(ValorantApi):
    pass



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class GameFeatureOverride:
    feature_name: str
    state: bool

    def __init__(self, data: dict) -> None:
        self.feature_name = data.get("featureName")
        self.state = data.get("state")


@dataclass
class GameRuleBoolOverride:
    rule_name: str
    state: bool

    def __init__(self, data: dict) -> None:
        self.rule_name = data.get("ruleName")
        self.state = data.get("state")


@dataclass
class GameMode:
    uuid: str
    display_name: Optional[str]
    duration: Optional[str]
    allows_match_timeouts: bool
    is_team_voice_allowed: bool
    is_minimap_hidden: bool
    orb_count: int
    team_roles: Optional[List[str]]
    game_feature_overrides: Optional[List[GameFeatureOverride]]
    game_rule_bool_overrides: Optional[List[GameRuleBoolOverride]]
    display_icon: Optional[str]
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.duration = data.get("duration")
        self.a = data.get("allowsMatchTimeouts")
        self.is_team_voice_allowed = data.get("isTeamVoiceAllowed")
        self.is_minimap_hidden = data.get("isMinimapHidden")
        self.orb_count = data.get("orbCount")
        self.team_roles = data.get("teamRoles")
        self.game_feature_overrides = data.get("gameFeatureOverrides")
        self.display_icon = data.get("displayIcon")
        self.asset_path = data.get("assetPath")
        self.raw_data = data


@dataclass
class Equippable:
    uuid: str
    display_name: str
    category: str
    display_icon: str
    kill_stream_icon: str
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict):
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.category = data.get("category")
        self.display_icon = data.get("displayIcon")
        self.kill_stream_icon = data.get("killStreamIcon")
        self.asset_path = data.get("assetPath")
        self.raw_data = data



from dataclasses import dataclass
from typing import List



@dataclass
class Location:
    x: float
    y: float

    def __init__(self, data: dict) -> None:
        self.x = data.get('x')
        self.y = data.get('y')


@dataclass
class Callout:
    region_name: str
    super_region_name: str
    location: Location

    def __init__(self, data: dict) -> None:
        self.region_name = data.get("regionName")
        self.super_region_name = data.get("superRegionName")
        self.location = Location(data.get("location"))


@dataclass
class Map:
    uuid: str
    display_name: str
    coordinates: str
    display_icon: str
    list_view_icon: str
    splash: str
    asset_path: str
    map_url: str
    x_multiplier: float
    y_multiplier: float
    x_scalar_to_add: float
    y_scalar_to_add: float
    callouts: List[Callout]
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.coordinates = data.get("coordinates")
        self.display_icon = data.get("displayIcon")
        self.list_view_icon = data.get("listViewIcon")
        self.splash = data.get("splash")
        self.asset_path = data.get("assetPath")
        self.map_url = data.get("mapUrl")
        self.x_multiplier = data.get("xMultiplier")
        self.y_multiplier = data.get("yMultiplier")
        self.x_scalar_to_add = data.get("xScalarToAdd")
        self.y_scalar_to_add = data.get("yScalarToAdd")
        self.callouts = [Callout(x) for x in data.get("callouts")] if data.get("callouts") else []
        self.raw_data = data



from dataclasses import dataclass
from typing import Optional


@dataclass
class Bundle:
    uuid: str
    display_name: str
    display_name_sub_text: Optional[str]
    description: str
    display_icon: str
    display_icon2: str
    vertical_promo_image: Optional[str]
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.display_name_sub_text = data.get("displayNameSubText")
        self.description = data.get("description")
        self.display_icon = data.get("displayIcon")
        self.display_icon2 = data.get("displayIcon2")
        self.vertical_promo_image = data.get("verticalPromoImage")
        self.asset_path = data.get("assetPath")
        self.raw_data = data




class Endpoints:
    BASE = "https://valorant-api.com/v1/"
    Agents = BASE + "agents"
    Buddies = BASE + "buddies"
    BuddyLevels = Buddies + "/levels"
    Bundles = BASE + "bundles"
    Ceremonies = BASE + "ceremonies"
    ContentTiers = BASE + "contenttiers"
    Currencies = BASE + "currencies"
    GameMode = BASE + "gamemodes"
    GamemodeEquippables = GameMode + "/equippables"
    Maps = BASE + "maps"
    PlayerCards = BASE + "playercards"
    PlayerTitles = BASE + "playertitles"
    Seasons = BASE + "seasons"
    Sprays = BASE + "sprays"
    SprayLevels = Sprays + "/levels"
    Themes = BASE + "themes"
    Weapons = BASE + "weapons"
    Version = BASE + "version"
    CompetitiveTiers = BASE + "competitivetiers"
    Competitive = Seasons + "/competitive"
    WeaponSkins = Weapons + "/skins"
    WeaponSkinChromas = Weapons + "/skinchromas"
    WeaponSkinLevels = Weapons + "/skinlevels"
    Contracts = BASE + "contracts"
    Events = BASE + "events"
    Gears = BASE + "gear"
    LevelBorders = BASE + "levelborders"



from dataclasses import dataclass


@dataclass
class ContentTier:
    uuid: str
    dev_name: str
    highlight_color: str
    display_icon: str
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict):
        self.uuid = data.get("uuid")
        self.dev_name = data.get("devName")
        self.highlight_color = data.get("highlightColor")
        self.display_icon = data.get("displayIcon")
        self.asset_path = data.get("assetPath")
        self.raw_data = data



from dataclasses import dataclass
from typing import Optional


@dataclass
class PlayerCard:
    uuid: str
    display_name: str
    is_hidden_if_not_owned: bool
    theme_uuid: Optional[str]
    display_icon: str
    small_art: str
    wide_art: str
    large_art: str
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.is_hidden_if_not_owned = data.get("isHiddenIfNotOwned")
        self.theme_uuid = data.get("themeUuid")
        self.display_icon = data.get("displayIcon")
        self.small_art = data.get("smallArt")
        self.wide_art = data.get("wideArt")
        self.large_art = data.get("largeArt")
        self.asset_path = data.get("assetPath")
        self.raw_data = data



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Tier:
    tier: int
    tier_name: str
    division: str
    division_name: str
    background_color: str
    color: str
    small_icon: Optional[str]
    large_icon: Optional[str]
    rank_triangle_down_icon: Optional[str]
    rank_triangle_up_icon: Optional[str]

    def __init__(self, data: dict) -> None:
        self.tier = data.get("tier")
        self.tier_name = data.get("tierName")
        self.division = data.get("division")
        self.division_name = data.get("divisionName")
        self.background_color = data.get("backgroundColor")
        self.color = data.get("color")
        self.small_icon = data.get("smallIcon")
        self.large_icon = data.get("largeIcon")
        self.rank_triangle_down_icon = data.get("rankTriangleDownIcon")
        self.rank_triangle_up_icon = data.get("rankTriangleUpIcon")


@dataclass
class CompetitiveTier:
    uuid: str
    asset_object_name: str
    tiers: Optional[List[Tier]]
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.asset_object_name = data.get("assetObjectName")
        self.tiers = [Tier(x) for x in data.get("tiers")] if data.get("tiers") is not None else None
        self.asset_path = data.get("assetPath")



from dataclasses import dataclass
from typing import Union


@dataclass
class Currency:
    uuid: str
    display_name: str
    display_name_singular: str
    display_icon: str
    large_icon: Union[str, None]
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict):
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.display_name_singular = data.get("displayNameSingular")
        self.display_icon = data.get("displayIcon")
        self.large_icon = data.get("largeIcon")
        self.asset_path = data.get("assetPath")
        self.raw_data = data



from dataclasses import dataclass
from typing import Union


@dataclass
class Theme:
    uuid: str
    display_name: str
    display_icon: Union[str, None]
    store_featured_image: Union[str, None]
    asset_path: str
    raw_data: dict

    def __init__(self, obj: dict) -> None:
        self.uuid = obj.get("uuid")
        self.display_name = obj.get("displayName")
        self.display_icon = obj.get("displayIcon")
        self.store_featured_image = obj.get("storeFeaturedImage")
        self.asset_path = obj.get("assetPath")
        self.raw_data = obj



from dataclasses import dataclass


@dataclass
class Ceremony:
    uuid: str
    display_name: str
    asset_path: str

    def __init__(self, data: dict) -> None:
        self.uuid = data.get('uuid')
        self.display_name = data.get('displayName')
        self.asset_path = data.get('assetPath')



from dataclasses import dataclass


@dataclass
class LevelBorder:
    uuid: str
    starting_level: int
    level_number_appearance: str
    small_player_card_appearance: str
    asset_path: str

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.starting_level = data.get("startingLevel")
        self.level_number_appearance = data.get("levelNumberAppearance")
        self.small_player_card_appearance = data.get("smallPlayerCardAppearance")
        self.asset_path = data.get("assetPath")



from dataclasses import dataclass
from typing import Optional, List


@dataclass
class GridPosition:
    row: int
    column: int

    def __init__(self, data: dict) -> None:
        self.row = data.get("row")
        self.column = data.get("column")


@dataclass
class ShopData:
    cost: int
    category: str
    category_text: str
    grid_position: Optional[GridPosition]
    image: str
    new_image: str
    new_image2: str
    asset_path: str

    def __init__(self, data: dict):
        self.cost = data.get("cost")
        self.category = data.get("category")
        self.category_text = data.get("categoryText")
        self.grid_position = GridPosition(data.get("gridPosition")) if data.get("gridPosition") is not None else None
        self.image = data.get("image")
        self.new_image = data.get("newImage")
        self.new_image2 = data.get("newImage2")
        self.asset_path = data.get("assetPath")


@dataclass
class Chroma:
    uuid: str
    display_name: str
    display_icon: str
    full_render: str
    swatch: Optional[str]
    streamed_video: Optional[str]
    asset_path: str

    def __init__(self, data: dict):
        self.raw_data = data
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.display_icon = data.get("displayIcon")
        self.full_render = data.get("fullRender")
        self.swatch = data.get("swatch")
        self.streamed_video = data.get("streamedVideo")
        self.asset_path = data.get("assetPath")


@dataclass
class Level:
    uuid: str
    display_name: str
    level_item: Optional[str]
    display_icon: str
    streamed_video: Optional[str]
    asset_path: str

    def __init__(self, data: dict) -> None:
        self.raw_data = data
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.level_item = data.get("levelItem")
        self.display_icon = data.get("displayIcon")
        self.streamed_video = data.get("streamedVideo")
        self.asset_path = data.get("assetPath")


@dataclass
class Skin:
    uuid: str
    display_name: str
    theme_uuid: str
    content_tier_uuid: Optional[str]
    wallpaper: Optional[str]
    display_icon: str
    asset_path: str
    chromas: List[Optional[Chroma]]
    levels: List[Optional[Level]]

    def __init__(self, data: dict):
        self.raw_data = data
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.theme_uuid = data.get("themeUuid")
        self.content_tier_uuid = data.get("contentTierUuid")
        self.wallpaper = data.get("wallpaper")
        self.display_icon = data.get("displayIcon")
        self.asset_path = data.get("assetPath")
        self.chromas = [Chroma(x) for x in data.get("chromas") if x is not None]
        self.levels = [Level(x) for x in data.get("levels") if x is not None]


@dataclass
class AdsStats:
    zoom_multiplier: float
    fire_rate: float
    run_speed_multiplier: float
    burst_count: int
    first_bullet_accuracy: float

    def __init__(self, data: dict):
        self.zoom_multiplier = data.get("zoomMultiplier")
        self.fire_rate = data.get("fireRate")
        self.run_speed_multiplier = data.get("runSpeedMultiplier")
        self.burst_count = data.get("burstCount")
        self.first_bullet_accuracy = data.get("firstBulletAccuracy")


@dataclass
class DamageRange:
    range_start_meters: int
    range_end_meters: int
    head_damage: float
    body_damage: int
    leg_damage: float

    def __init__(self, data: dict) -> None:
        self.range_start_meters = data.get("rangeStartMeters")
        self.range_end_meters = data.get("rangeEndMeters")
        self.head_damage = data.get("headDamage")
        self.body_damage = data.get("bodyDamage")
        self.leg_damage = data.get("legDamage")


@dataclass
class AirBurstStats:
    shotgun_pellet_count: int
    burst_distance: float

    def __init__(self, data: dict):
        self.shotgun_pellet_count = data.get("shotgunPelletCount")
        self.burst_distance = data.get("burstDistance")


@dataclass
class AltShotgunStats:
    shotgun_pellet_count: int
    burst_rate: float

    def __init__(self, data: dict) -> None:
        self.shotgun_pellet_count = data.get("shotgunPelletCount")
        self.burst_rate = data.get("burstRate")


@dataclass
class WeaponStats:
    fire_rate: int
    magazine_size: int
    run_speed_multiplier: float
    equip_time_seconds: float
    reload_time_seconds: int
    first_bullet_accuracy: float
    shotgun_pellet_count: int
    wall_penetration: str
    feature: str
    fire_mode: Optional[str]
    alt_fire_type: str
    ads_stats: Optional[AdsStats]
    alt_shotgun_stats: AltShotgunStats
    air_burst_stats: AirBurstStats
    damage_ranges: List[DamageRange]

    def __init__(self, data: dict) -> None:
        self.fire_rate = data.get("fireRate")
        self.magazine_size = data.get("magazineSize")
        self.run_speed_multiplier = data.get("runSpeedMultiplier")
        self.equip_time_seconds = data.get("equipTimeSeconds")
        self.reload_time_seconds = data.get("reloadTimeSeconds")
        self.first_bullet_accuracy = data.get("firstBulletAccuracy")
        self.shotgun_pellet_count = data.get("shotgunPelletCount")
        self.wall_penetration = data.get("wallPenetration")
        self.feature = data.get("feature")
        self.fire_mode = data.get("fireMode")
        self.alt_fire_type = data.get("altFireType")
        self.ads_stats = AdsStats(data.get("adsStats")) if data.get("adsStats") is not None else None
        self.alt_shotgun_stats = data.get("altShotgunStats")
        self.air_burst_stats = data.get("airBurstStats")
        self.damage_ranges = [DamageRange(x) for x in data.get("damageRanges")]


@dataclass
class Weapon:
    uuid: str
    display_name: str
    category: str
    default_skin_uuid: str
    display_icon: str
    kill_stream_icon: str
    asset_path: str
    weapon_stats: Optional[WeaponStats]
    shop_data: Optional[ShopData]
    skins: Optional[List[Skin]]
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.category = data.get("category")
        self.default_skin_uuid = data.get("defaultSkinUuid")
        self.display_icon = data.get("displayIcon")
        self.kill_stream_icon = data.get("killStreamIcon")
        self.asset_path = data.get("assetPath")
        self.weapon_stats = WeaponStats(data.get("weaponStats")) if data.get("weaponStats") is not None else None
        self.shop_data = ShopData(data.get("shopData")) if data.get("shopData") is not None else None
        self.skins = [Skin(x) for x in data.get("skins")] if data.get("skins") is not None else None
        self.raw_data = data



from dataclasses import dataclass
from datetime import datetime

import dateutil.parser


@dataclass
class Version:
    manifest_id: str
    branch: str
    version: str
    build_version: str
    build_date: datetime
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.manifest_id = data.get("manifestId")
        self.branch = data.get("branch")
        self.version = data.get("version")
        self.build_version = data.get("buildVersion")
        self.build_date = dateutil.parser.parse(data.get("buildDate"))
        self.raw_data = data



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Ability:
    slot: str
    display_name: str
    description: str
    display_icon: str

    def __init__(self, data: dict) -> None:
        self.slot = data.get("slot")
        self.display_name = data.get("displayName")
        self.description = data.get("description")
        self.display_icon = data.get("displayIcon")


@dataclass
class Role:
    uuid: str
    display_name: str
    description: str
    display_icon: str
    asset_path: str

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.description = data.get("description")
        self.display_icon = data.get("displayIcon")
        self.asset_path = data.get("assetPath")


@dataclass
class Media:
    id: str
    wwise: str
    wave: str

    def __init__(self, data: dict) -> None:
        self.id = data.get("id")
        self.wwise = data.get("wwise")
        self.wave = data.get("wave")


@dataclass
class VoiceLine:
    min_duration: float
    max_duration: float
    media_list: List[Media]

    def __init__(self, data: dict) -> None:
        self.min_duration = data.get("minDuration")
        self.max_duration = data.get("maxDuration")
        self.media_list = [Media(x) for x in data.get("mediaList")]

@dataclass
class Agent:
    uuid: str
    display_name: str
    description: str
    developer_name: str
    character_tags: Optional[list]
    display_icon: str
    display_icon_small: str
    bust_portrait: str
    full_portrait: str
    full_portraitV2: str
    kill_feed_portrait: str
    background: str
    background_gradient_colors: List[str]
    asset_path: str
    is_full_portrait_right_facing: bool
    is_playable_character: bool
    is_available_for_test: bool
    role: Optional[Role]
    abilities: Optional[List[Ability]]
    voice_line: Optional[VoiceLine]
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.description = data.get("description")
        self.developer_name = data.get("developerName")
        self.character_tags = data.get("characterTags")
        self.display_icon = data.get("displayIcon")
        self.display_icon_small = data.get("displayIconSmall")
        self.bust_portrait = data.get("bustPortrait")
        self.full_portrait = data.get("fullPortrait")
        self.full_portraitV2 = data.get("fullPortraitV2")
        self.kill_feed_portrait = data.get("killfeedPortrait")
        self.background = data.get("background")
        self.background_gradient_colors = data.get("backgroundGradientColors")
        self.asset_path = data.get("assetPath")
        self.is_full_portrait_right_facing = data.get("isFullPortraitRightFacing")
        self.is_playable_character = data.get("isPlayableCharacter")
        self.is_available_for_test = data.get("isAvailableForTest")
        self.role = Role(data.get("role")) if data.get("role") is not None else None
        self.abilities = [Ability(x) for x in data.get("abilities")] if data.get("abilities") is not None else None
        self.voice_line = VoiceLine(data.get("voiceLine")) if data.get("voiceLine") is not None else None
        self.raw_data = data



from .ceremonies import Ceremony
from .gear import Gear
from .events import Event
from .contracts import Contract
from .competitive import Competitive
from .competitivetiers import CompetitiveTier
from .sprays import Spray
from .sprays import Level as SprayLevel
from .agents import Agent
from .base_list import BaseList
from .buddies import Buddy
from .buddies import Level as BuddyLevel
from .bundles import Bundle
from .contenttiers import ContentTier
from .currencies import Currency
from .endpoints import Endpoints
from .gamemodes import GameMode, Equippable
from .maps import Map
from .playercards import PlayerCard
from .playertitles import PlayerTitle
from .seasons import Season
from .themes import Theme
from .version import Version
from .weapons import Chroma, Skin, Weapon, Level
from .levelborders import LevelBorder

from .httpclient import AsyncClient, SyncClient


class SyncValorantApi:
    """
Synchronous valorant-api wrapper
    """

    headers: dict
    client: SyncClient
    params: dict = {}

    def __init__(self, headers=None, language: str = "en-US"):
        if headers is None:
            headers = {}
        self.params["language"] = language
        self.headers = headers
        self.client = SyncClient(headers, params=self.params)

    def get_agents(self, *args, **kwargs) -> BaseList[Agent]:
        data = self.client.get(Endpoints.Agents, *args, **kwargs)
        return BaseList([Agent(x) for x in data])

    def search_agents_by_uuid(self, uuid: str, *args, **kwargs) -> Agent:
        data = self.client.get(f"{Endpoints.Agents}/{uuid}", *args, **kwargs)
        return Agent(data)

    def get_buddies(self, *args, **kwargs) -> BaseList[Buddy]:
        data = self.client.get(Endpoints.Buddies, *args, **kwargs)
        return BaseList([Buddy(x) for x in data])

    def search_buddies_by_uuid(self, uuid: str, *args, **kwargs) -> Buddy:
        data = self.client.get(f"{Endpoints.Buddies}/{uuid}", *args, **kwargs)
        return Buddy(data)

    def get_bundles(self, *args, **kwargs) -> BaseList[Bundle]:
        data = self.client.get(Endpoints.Bundles, *args, **kwargs)
        return BaseList([Bundle(x) for x in data])

    def search_bundles_by_uuid(self, uuid: str, *args, **kwargs) -> Bundle:
        data = self.client.get(f"{Endpoints.Bundles}/{uuid}", *args, **kwargs)
        return Bundle(data)

    def get_contenttiers(self, *args, **kwargs) -> BaseList[ContentTier]:
        data = self.client.get(Endpoints.ContentTiers, *args, **kwargs)
        return BaseList([ContentTier(x) for x in data])

    def search_contenttier_by_uuid(self, uuid: str, *args, **kwargs) -> ContentTier:
        data = self.client.get(f"{Endpoints.ContentTiers}/{uuid}", *args, **kwargs)
        return ContentTier(data)

    def get_currencies(self, *args, **kwargs) -> BaseList[Currency]:
        data = self.client.get(Endpoints.Currencies, *args, **kwargs)
        return BaseList([Currency(x) for x in data])

    def search_currencies_by_uuid(self, uuid: str, *args, **kwargs) -> ContentTier:
        data = self.client.get(f"{Endpoints.Currencies}/{uuid}", *args, **kwargs)
        return ContentTier(data)

    def get_ceremonies(self, *args, **kwargs) -> BaseList[Ceremony]:
        data = self.client.get(Endpoints.Ceremonies, *args, **kwargs)
        return BaseList([Ceremony(x) for x in data])

    def search_ceremonies_by_uuid(self, uuid: str, *args, **kwargs) -> Ceremony:
        data = self.client.get(f"{Endpoints.Ceremonies}/{uuid}", *args, **kwargs)
        return Ceremony(data)

    def get_gamemodes(self, *args, **kwargs) -> BaseList[GameMode]:
        data = self.client.get(Endpoints.GameMode, *args, **kwargs)
        return BaseList([GameMode(x) for x in data])

    def search_gamemodes_by_uuid(self, uuid: str, *args, **kwargs) -> GameMode:
        data = self.client.get(f"{Endpoints.GameMode}/{uuid}", *args, **kwargs)
        return GameMode(data)

    def get_gamemode_equippables(self, *args, **kwargs) -> BaseList[Equippable]:
        data = self.client.get(Endpoints.GamemodeEquippables, *args, **kwargs)
        return BaseList([Equippable(x) for x in data])

    def search_gamemode_equippables_by_uuid(self, uuid: str, *args, **kwargs) -> Equippable:
        data = self.client.get(f"{Endpoints.GamemodeEquippables}/{uuid}", *args, **kwargs)
        return Equippable(data)

    def get_maps(self, *args, **kwargs) -> BaseList[Map]:
        data = self.client.get(Endpoints.Maps, *args, **kwargs)
        return BaseList([Map(x) for x in data])

    def search_maps_by_uuid(self, uuid: str, *args, **kwargs) -> Map:
        data = self.client.get(f"{Endpoints.Maps}/{uuid}", *args, **kwargs)
        return Map(data)

    def get_playercards(self, *args, **kwargs) -> BaseList[PlayerCard]:
        data = self.client.get(Endpoints.PlayerCards, *args, **kwargs)
        return BaseList([PlayerCard(x) for x in data])

    def search_playercards_by_uuid(self, uuid: str, *args, **kwargs) -> PlayerCard:
        data = self.client.get(f"{Endpoints.PlayerCards}/{uuid}", *args, **kwargs)
        return PlayerCard(data)

    def get_playertitles(self, *args, **kwargs) -> BaseList[PlayerTitle]:
        data = self.client.get(Endpoints.PlayerTitles, *args, **kwargs)
        return BaseList([PlayerTitle(x) for x in data])

    def search_playertitles_by_uuid(self, uuid: str, *args, **kwargs) -> PlayerTitle:
        data = self.client.get(f"{Endpoints.PlayerTitles}/{uuid}", *args, **kwargs)
        return PlayerTitle(data)

    def get_seasons(self, *args, **kwargs) -> BaseList[Season]:
        data = self.client.get(Endpoints.Seasons, *args, **kwargs)
        return BaseList([Season(x) for x in data])

    def search_seasons_by_uuid(self, uuid: str, *args, **kwargs) -> Season:
        data = self.client.get(f"{Endpoints.Seasons}/{uuid}", *args, **kwargs)
        return Season(data)

    def get_themes(self, *args, **kwargs) -> BaseList[Theme]:
        data = self.client.get(Endpoints.Themes, *args, **kwargs)
        return BaseList([Theme(x) for x in data])

    def search_themes_by_uuid(self, uuid: str, *args, **kwargs) -> Theme:
        data = self.client.get(f"{Endpoints.Themes}/{uuid}", *args, **kwargs)
        return Theme(data)

    def get_weapons(self, *args, **kwargs) -> BaseList[Weapon]:
        data = self.client.get(Endpoints.Weapons, *args, **kwargs)
        return BaseList([Weapon(x) for x in data])

    def search_weapons_by_uuid(self, uuid: str, *args, **kwargs) -> Weapon:
        data = self.client.get(f"{Endpoints.Weapons}/{uuid}", *args, **kwargs)
        return Weapon(data)

    def get_sprays(self, *args, **kwargs) -> BaseList[Spray]:
        data = self.client.get(Endpoints.Sprays, *args, **kwargs)
        return BaseList([Spray(x) for x in data])

    def search_sprays_by_uuid(self, uuid: str, *args, **kwargs) -> Spray:
        data = self.client.get(f"{Endpoints.Sprays}/{uuid}", *args, **kwargs)
        return Spray(data)

    def get_version(self, *args, **kwargs) -> Version:
        data = self.client.get(f"{Endpoints.Version}", *args, **kwargs)
        return Version(data)

    def get_competitivetiers(self, *args, **kwargs) -> BaseList[CompetitiveTier]:
        data = self.client.get(Endpoints.CompetitiveTiers, *args, **kwargs)
        return BaseList([CompetitiveTier(x) for x in data])

    def search_competitivetiers_by_uuid(self, uuid: str, *args, **kwargs) -> CompetitiveTier:
        data = self.client.get(f"{Endpoints.CompetitiveTiers}/{uuid}", *args, **kwargs)
        return CompetitiveTier(data)

    def get_competitive(self, *args, **kwargs) -> BaseList[Competitive]:
        data = self.client.get(Endpoints.Competitive, *args, **kwargs)
        return BaseList([Competitive(x) for x in data])
    
    def search_competitive_by_uuid(self, uuid: str, *args, **kwargs) -> BaseList[Competitive]:
        data = self.client.get(f"{Endpoints.Competitive}/{uuid}", *args, **kwargs)
        return Competitive(data)

    def get_weapon_skins(self, *args, **kwargs) -> BaseList[Skin]:
        data = self.client.get(Endpoints.WeaponSkins, *args, **kwargs)
        return BaseList([Skin(x) for x in data])

    def search_weapon_skins_by_uuid(self, uuid: str, *args, **kwargs):
        data = self.client.get(f"{Endpoints.WeaponSkins}/{uuid}", *args, **kwargs)
        return Skin(data)

    def get_weapons_levels(self, *args, **kwargs) -> BaseList[Level]:
        data = self.client.get(Endpoints.WeaponSkinLevels, *args, **kwargs)
        return BaseList([Level(x) for x in data])

    def search_weapon_levels_by_uuid(self, uuid: str, *args, **kwargs) -> Level:
        data = self.client.get(f"{Endpoints.WeaponSkinLevels}/{uuid}", *args, **kwargs)
        return Level(data)

    def get_weapon_chromas(self, *args, **kwargs) -> BaseList[Chroma]:
        data = self.client.get(Endpoints.WeaponSkinChromas, *args, **kwargs)
        return BaseList([Chroma(x) for x in data])

    def search_weapon_chromas_by_uuid(self, uuid: str, *args, **kwargs) -> Chroma:
        data = self.client.get(f"{Endpoints.WeaponSkinChromas}/{uuid}", *args, **kwargs)
        return Chroma(data)

    def get_contracts(self, *args, **kwargs) -> BaseList[Contract]:
        data = self.client.get(Endpoints.Contracts, *args, **kwargs)
        return BaseList([Contract(x) for x in data])

    def search_contracts_by_uuid(self, uuid: str, *args, **kwargs) -> Chroma:
        data = self.client.get(f"{Endpoints.Contracts}/{uuid}", *args, **kwargs)
        return Contract(data)

    def get_events(self, *args, **kwargs) -> BaseList[Event]:
        data = self.client.get(Endpoints.Events, *args, **kwargs)
        return BaseList([Event(x) for x in data])

    def search_events_by_uuid(self, uuid: str, *args, **kwargs) -> Event:
        data = self.client.get(f"{Endpoints.Events}/{uuid}", *args, **kwargs)
        return Contract(data)

    def get_gears(self, *args, **kwargs) -> BaseList[Gear]:
        data = self.client.get(Endpoints.Gears, *args, **kwargs)
        return BaseList([Gear(x) for x in data])

    def search_gears_by_uuid(self, uuid: str, *args, **kwargs) -> Gear:
        data = self.client.get(f"{Endpoints.Gears}/{uuid}", *args, **kwargs)
        return Gear(data)

    def get_buddy_levels(self, *args, **kwargs) -> BaseList[BuddyLevel]:
        data = self.client.get(Endpoints.BuddyLevels, *args, **kwargs)
        return BaseList([BuddyLevel(x) for x in data])

    def search_buddy_levels_by_uuid(self, uuid: str, *args, **kwargs) -> BuddyLevel:
        data = self.client.get(f"{Endpoints.BuddyLevels}/{uuid}", *args, **kwargs)
        return BuddyLevel(data)

    def get_spray_levels(self, *args, **kwargs) -> BaseList[SprayLevel]:
        data = self.client.get(Endpoints.SprayLevels, *args, **kwargs)
        return BaseList([SprayLevel(x) for x in data])

    def search_spray_levels_by_uuid(self, uuid: str, *args, **kwargs) -> SprayLevel:
        data = self.client.get(f"{Endpoints.SprayLevels}/{uuid}", *args, **kwargs)
        return SprayLevel(data)

    def get_level_borders(self, *args, **kwargs) -> BaseList[LevelBorder]:
        data = self.client.get(Endpoints.LevelBorders, *args, **kwargs)
        return BaseList([LevelBorder(x) for x in data])

    def search_level_borders_by_uuid(self, uuid: str, *args, **kwargs) -> LevelBorder:
        data = self.client.get(f"{Endpoints.LevelBorders}/{uuid}", *args, **kwargs)
        return LevelBorder(data)

class AsyncValorantApi:
    """
Asynchronous valorant-api wrapper
    """

    headers: dict
    client: AsyncClient
    params: dict = {}

    def __init__(self, headers=None, language: str = "en-US"):
        if headers is None:
            headers = {}
        self.params["language"] = language
        self.headers = headers
        self.client = AsyncClient(headers, params=self.params)

    async def get_agents(self, *args, **kwargs) -> BaseList[Agent]:
        data = await self.client.get(Endpoints.Agents, *args, **kwargs)
        return BaseList([Agent(x) for x in data])

    async def search_agents_by_uuid(self, uuid: str, *args, **kwargs) -> Agent:
        data = await self.client.get(f"{Endpoints.Agents}/{uuid}", *args, **kwargs)
        return Agent(data)

    async def get_buddies(self, *args, **kwargs) -> BaseList[Buddy]:
        data = await self.client.get(Endpoints.Buddies, *args, **kwargs)
        return BaseList([Buddy(x) for x in data])

    async def search_buddies_by_uuid(self, uuid: str, *args, **kwargs) -> Buddy:
        data = await self.client.get(f"{Endpoints.Buddies}/{uuid}", *args, **kwargs)
        return Buddy(data)

    async def get_bundles(self, *args, **kwargs) -> BaseList[Bundle]:
        data = await self.client.get(Endpoints.Bundles, *args, **kwargs)
        return BaseList([Bundle(x) for x in data])

    async def search_bundles_by_uuid(self, uuid: str, *args, **kwargs) -> Bundle:
        data = await self.client.get(f"{Endpoints.Bundles}/{uuid}", *args, **kwargs)
        return Bundle(data)

    async def get_contenttiers(self, *args, **kwargs) -> BaseList[ContentTier]:
        data = await self.client.get(Endpoints.ContentTiers, *args, **kwargs)
        return BaseList([ContentTier(x) for x in data])

    async def search_contenttier_by_uuid(self, uuid: str, *args, **kwargs) -> ContentTier:
        data = await self.client.get(f"{Endpoints.ContentTiers}/{uuid}", *args, **kwargs)
        return ContentTier(data)

    async def get_currencies(self, *args, **kwargs) -> BaseList[Currency]:
        data = await self.client.get(Endpoints.Currencies, *args, **kwargs)
        return BaseList([Currency(x) for x in data])

    async def search_currencies_by_uuid(self, uuid: str, *args, **kwargs) -> ContentTier:
        data = await self.client.get(f"{Endpoints.Currencies}/{uuid}", *args, **kwargs)
        return ContentTier(data)

    async def get_ceremonies(self, *args, **kwargs) -> BaseList[Ceremony]:
        data = await self.client.get(Endpoints.Ceremonies, *args, **kwargs)
        return BaseList([Ceremony(x) for x in data])

    async def search_ceremonies_by_uuid(self, uuid: str, *args, **kwargs) -> Ceremony:
        data = await self.client.get(f"{Endpoints.Ceremonies}/{uuid}", *args, **kwargs)
        return Ceremony(data)

    async def get_gamemodes(self, *args, **kwargs) -> BaseList[GameMode]:
        data = await self.client.get(Endpoints.GameMode, *args, **kwargs)
        return BaseList([GameMode(x) for x in data])

    async def search_gamemodes_by_uuid(self, uuid: str, *args, **kwargs) -> GameMode:
        data = await self.client.get(f"{Endpoints.GameMode}/{uuid}", *args, **kwargs)
        return GameMode(data)

    async def get_gamemode_equippables(self, *args, **kwargs) -> BaseList[Equippable]:
        data = await self.client.get(Endpoints.GamemodeEquippables, *args, **kwargs)
        return BaseList([Equippable(x) for x in data])

    async def search_gamemode_equippables_by_uuid(self, uuid: str, *args, **kwargs) -> Equippable:
        data = await self.client.get(f"{Endpoints.GamemodeEquippables}/{uuid}", *args, **kwargs)
        return Equippable(data)

    async def get_maps(self, *args, **kwargs) -> BaseList[Map]:
        data = await self.client.get(Endpoints.Maps, *args, **kwargs)
        return BaseList([Map(x) for x in data])

    async def search_maps_by_uuid(self, uuid: str, *args, **kwargs) -> Map:
        data = await self.client.get(f"{Endpoints.Maps}/{uuid}", *args, **kwargs)
        return Map(data)

    async def get_playercards(self, *args, **kwargs) -> BaseList[PlayerCard]:
        data = await self.client.get(Endpoints.PlayerCards, *args, **kwargs)
        return BaseList([PlayerCard(x) for x in data])

    async def search_playercards_by_uuid(self, uuid: str, *args, **kwargs) -> PlayerCard:
        data = await self.client.get(f"{Endpoints.PlayerCards}/{uuid}", *args, **kwargs)
        return PlayerCard(data)

    async def get_playertitles(self, *args, **kwargs) -> BaseList[PlayerTitle]:
        data = await self.client.get(Endpoints.PlayerTitles, *args, **kwargs)
        return BaseList([PlayerTitle(x) for x in data])

    async def search_playertitles_by_uuid(self, uuid: str, *args, **kwargs) -> PlayerTitle:
        data = await self.client.get(f"{Endpoints.PlayerTitles}/{uuid}", *args, **kwargs)
        return PlayerTitle(data)

    async def get_seasons(self, *args, **kwargs) -> BaseList[Season]:
        data = await self.client.get(Endpoints.Seasons, *args, **kwargs)
        return BaseList([Season(x) for x in data])

    async def search_seasons_by_uuid(self, uuid: str, *args, **kwargs) -> Season:
        data = await self.client.get(f"{Endpoints.Seasons}/{uuid}", *args, **kwargs)
        return Season(data)

    async def get_themes(self, *args, **kwargs) -> BaseList[Theme]:
        data = await self.client.get(Endpoints.Themes, *args, **kwargs)
        return BaseList([Theme(x) for x in data])

    async def search_themes_by_uuid(self, uuid: str, *args, **kwargs) -> Theme:
        data = await self.client.get(f"{Endpoints.Themes}/{uuid}", *args, **kwargs)
        return Theme(data)

    async def get_weapons(self, *args, **kwargs) -> BaseList[Weapon]:
        data = await self.client.get(Endpoints.Weapons, *args, **kwargs)
        return BaseList([Weapon(x) for x in data])

    async def search_weapons_by_uuid(self, uuid: str, *args, **kwargs) -> Weapon:
        data = await self.client.get(f"{Endpoints.Weapons}/{uuid}", *args, **kwargs)
        return Weapon(data)

    async def get_sprays(self, *args, **kwargs) -> BaseList[Spray]:
        data = await self.client.get(Endpoints.Sprays, *args, **kwargs)
        return BaseList([Spray(x) for x in data])

    async def search_sprays_by_uuid(self, uuid: str, *args, **kwargs) -> Spray:
        data = await self.client.get(f"{Endpoints.Sprays}/{uuid}", *args, **kwargs)
        return Spray(data)

    async def get_version(self, *args, **kwargs) -> Version:
        data = await self.client.get(f"{Endpoints.Version}", *args, **kwargs)
        return Version(data)

    async def get_competitivetiers(self, *args, **kwargs) -> BaseList[CompetitiveTier]:
        data = await self.client.get(Endpoints.CompetitiveTiers, *args, **kwargs)
        return BaseList([CompetitiveTier(x) for x in data])

    async def search_competitivetiers_by_uuid(self, uuid: str, *args, **kwargs) -> CompetitiveTier:
        data = await self.client.get(f"{Endpoints.CompetitiveTiers}/{uuid}", *args, **kwargs)
        return CompetitiveTier(data)

    async def get_competitive(self, *args, **kwargs) -> BaseList[Competitive]:
        data = await self.client.get(Endpoints.Competitive, *args, **kwargs)
        return BaseList([Competitive(x) for x in data])

    async def search_competitive_by_uuid(self, uuid: str, *args, **kwargs) -> BaseList[Competitive]:
        data = await self.client.get(f"{Endpoints.Competitive}/{uuid}", *args, **kwargs)
        return Competitive(data)

    async def get_weapon_skins(self, *args, **kwargs) -> BaseList[Skin]:
        data = await self.client.get(Endpoints.WeaponSkins, *args, **kwargs)
        return BaseList([Skin(x) for x in data])

    async def search_weapon_skins_by_uuid(self, uuid: str, *args, **kwargs) -> Skin:
        data = await self.client.get(f"{Endpoints.WeaponSkins}/{uuid}", *args, **kwargs)
        return Skin(data)

    async def get_weapons_levels(self, *args, **kwargs) -> BaseList[Level]:
        data = await self.client.get(Endpoints.WeaponSkinLevels, *args, **kwargs)
        return BaseList([Level(x) for x in data])

    async def search_weapon_levels_by_uuid(self, uuid: str, *args, **kwargs) -> Level:
        data = await self.client.get(f"{Endpoints.WeaponSkinLevels}/{uuid}", *args, **kwargs)
        return Level(data)

    async def get_weapon_chromas(self, *args, **kwargs) -> BaseList[Chroma]:
        data = await self.client.get(Endpoints.WeaponSkinChromas, *args, **kwargs)
        return BaseList([Chroma(x) for x in data])

    async def search_weapon_chromas_by_uuid(self, uuid: str, *args, **kwargs) -> Chroma:
        data = await self.client.get(f"{Endpoints.WeaponSkinChromas}/{uuid}", *args, **kwargs)
        return Chroma(data)

    async def get_contracts(self, *args, **kwargs) -> BaseList[Contract]:
        data = await self.client.get(Endpoints.Contracts, *args, **kwargs)
        return BaseList([Contract(x) for x in data])

    async def search_contracts_by_uuid(self, uuid: str, *args, **kwargs) -> Chroma:
        data = await self.client.get(f"{Endpoints.Contracts}/{uuid}", *args, **kwargs)
        return Contract(data)

    async def get_events(self, *args, **kwargs) -> BaseList[Event]:
        data = await self.client.get(Endpoints.Events, *args, **kwargs)
        return BaseList([Event(x) for x in data])

    async def search_events_by_uuid(self, uuid: str, *args, **kwargs) -> Event:
        data = await self.client.get(f"{Endpoints.Events}/{uuid}", *args, **kwargs)
        return Contract(data)

    async def get_gears(self, *args, **kwargs) -> BaseList[Gear]:
        data = await self.client.get(Endpoints.Gears, *args, **kwargs)
        return BaseList([Gear(x) for x in data])

    async def search_gears_by_uuid(self, uuid: str, *args, **kwargs) -> Gear:
        data = await self.client.get(f"{Endpoints.Gears}/{uuid}", *args, **kwargs)
        return Gear(data)

    async def get_buddy_levels(self, *args, **kwargs) -> BaseList[BuddyLevel]:
        data = await self.client.get(Endpoints.BuddyLevels, *args, **kwargs)
        return BaseList([BuddyLevel(x) for x in data])

    async def search_buddy_levels_by_uuid(self, uuid: str, *args, **kwargs) -> BuddyLevel:
        data = await self.client.get(f"{Endpoints.BuddyLevels}/{uuid}", *args, **kwargs)
        return BuddyLevel(data)

    async def get_spray_levels(self, *args, **kwargs) -> BaseList[SprayLevel]:
        data = await self.client.get(Endpoints.SprayLevels, *args, **kwargs)
        return BaseList([SprayLevel(x) for x in data])

    async def search_spray_levels_by_uuid(self, uuid: str, *args, **kwargs) -> SprayLevel:
        data = await self.client.get(f"{Endpoints.SprayLevels}/{uuid}", *args, **kwargs)
        return SprayLevel(data)

    async def get_level_borders(self, *args, **kwargs) -> BaseList[LevelBorder]:
        data = await self.client.get(Endpoints.LevelBorders, *args, **kwargs)
        return BaseList([LevelBorder(x) for x in data])

    async def search_level_borders_by_uuid(self, uuid: str, *args, **kwargs) -> LevelBorder:
        data = await self.client.get(f"{Endpoints.LevelBorders}/{uuid}", *args, **kwargs)
        return LevelBorder(data)


from dataclasses import dataclass
from typing import Optional
from .utils import Mapper


@dataclass
class GridPosition(Mapper):
    row: Optional[int] = None
    column: Optional[int] = None

    def __init__(self):
        super().__init__()

@dataclass
class ShopData(Mapper):
    cost: Optional[int] = None
    category: Optional[str] = None
    category_text: Optional[str] = None
    grid_position: Optional[GridPosition] = None
    can_be_trashed: Optional[bool] = None
    image: Optional[str] = None
    new_image: Optional[str] = None
    new_image2: Optional[str] = None
    asset_path: Optional[str] = None

    def __init__(self):
        super().__init__()

@dataclass
class Gear(Mapper):
    uuid: str
    raw_data: dict
    display_name: Optional[str] = None
    description: Optional[str] = None
    display_icon: Optional[str] = None
    asset_path: Optional[str] = None
    shop_data: Optional[ShopData] = None

    def __init__(self, data: dict):
        self.raw_data = data
        self.map(childclasses={"shopData": ShopData, "gridPosition": GridPosition})



from dataclasses import dataclass
from typing import List, Optional
from .utils import Mapper


@dataclass
class Border(Mapper):
    uuid: str
    level: int
    wins_required: int
    display_icon: str
    small_icon: Optional[str]
    asset_path: str

    def __init__(self):
        super().__init__()


@dataclass
class Competitive(Mapper):
    uuid: str
    start_time: str
    end_time: str
    season_uuid: str
    competitive_tiers_uuid: str
    borders: List[Border]
    asset_path: str

    def __init__(self, data: dict) -> None:
        self.raw_data = data
        self.map(childclasses={"borders": Border})



from dataclasses import dataclass


@dataclass
class PlayerTitle:
    uuid: str
    display_name: str
    title_text: str  # same as displayName but without the " Title" suffix
    isb_hidden_if_not_owned: bool
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.title_text = data.get("titleText")
        self.isb_hidden_if_not_owned = data.get("isbHiddenIfNotOwned")
        self.asset_path = data.get("assetPath")
        self.raw_data = data

    def __str__(self):
        return str(self.raw_data)



from dataclasses import dataclass
from datetime import datetime
from typing import Union

import dateutil.parser


@dataclass
class Season:
    uuid: str
    display_name: str
    type: Union[str, None]
    start_time: datetime
    end_time: datetime
    parent_uuid: Union[str, None]
    asset_path: str
    raw_data: dict

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.type = data.get("type")
        self.start_time = dateutil.parser.parse(data.get("startTime"))
        self.end_time = dateutil.parser.parse(data.get("endTime"))
        self.parent_uuid = data.get("parentUuid")
        self.asset_path = data.get("assetPath")
        self.raw_data = data

    def __str__(self):
        return str(self.raw_data)



import asyncio
import atexit
import json
from http.client import HTTPException
import aiohttp
import requests
from aiohttp import ClientSession
from requests import Session
import copy

from .exceptions import *


def override_params(par1, par2):
    if par2 is not None:
        par1.update(par2)
    return par1

class SyncClient:
    headers: dict = {}
    params: dict = {}
    session: Session

    def __init__(self, headers=None, params=None) -> None:
        if params is None:
            self.params = {}
        else:
            self.params = params
        if headers is None:
            self.headers = {}
        else:
            self.headers = headers
            self.session = requests.Session()

    def get(self, url: str, **kwargs) -> dict:
        response = self.session.get(url, params=override_params(copy.copy(self.params), kwargs), headers=self.headers)

        try:
            data = response.json()
        except (json.JSONDecodeError):# , simplejson.JSONDecodeError):  # linux uses simplejson??
            data = {'error': response.text}

        if response.status_code == 200:
            return data["data"]
        elif response.status_code == 400:
            raise InvalidOrMissingParameter(data["error"])
        elif response.status_code == 404:
            raise NotFound(data["error"])
        else:
            raise HTTPException(f'{data.get("error", "An error unknown occurred")}, status code {response.status_code}')


class AsyncClient:
    headers: dict = {}
    params: dict = {}
    session: ClientSession

    def __init__(self, headers=None, params=None) -> None:
        if params is None:
            self.params = {}
        else:
            self.params = params
        if headers is None:
            self.headers = {}
        else:
            self.headers = headers
        self.session = aiohttp.ClientSession()
        atexit.register(self.close)

    def close(self):
        asyncio.run(self.session.close())

    async def get(self, url: str, **kwargs) -> dict:
        async with self.session.get(url, params=override_params(copy.copy(self.params), kwargs), headers=self.headers) as response:
            try:
                data = await response.json()
            except aiohttp.ContentTypeError:
                data = {'error': response.text}

            if response.status == 200:
                return data["data"]
            elif response.status == 400:
                raise InvalidOrMissingParameter(data["error"])
            elif response.status == 404:
                raise NotFound(data["error"])
            else:
                raise HTTPException(
                    f'{data.get("error", "An error unknown occurred")}, status code {response.status}')



from typing import List, Union, Optional, Any, Generic, TypeVar

T = TypeVar('T')


class BaseList(list, Generic[T]):
    _list: List[T]

    def __init__(self, data: List[T]):
        self._list = data
        super(BaseList, self).__init__(self._list)

    def __iter__(self):
        for x in self._list:
            yield x

    def find_first(self, **kwargs) -> Optional[T]:
        """returns None if not found else first matching result."""
        kwargs = dict((k.lower(), v.lower()) for k, v in kwargs.items())  # to lower case key/value
        searchkeys = [x for x in list(kwargs.keys())]
        x: dict[str, str]
        for datacls in self._list:
            x = datacls.raw_data
            for key, val in x.items():
                key = key.lower()
                if isinstance(val, str):
                    val = val.lower()

                if key in searchkeys:
                    if val == kwargs[key].lower():
                        return datacls
        return None

    def find_where(self, **kwargs) -> List[T]:
        """returns empty list if not found else list of matching result."""
        result = []
        kwargs = dict((k.lower(), v.lower()) for k, v in kwargs.items())  # to lower case key/value
        searchkeys = [x for x in list(kwargs.keys())]
        x: dict[str, str]
        for datacls in self._list:
            x = datacls.raw_data
            for key, val in x.items():
                key = key.lower()
                if isinstance(val, str):
                    val = val.lower()

                if key in searchkeys:
                    if val == kwargs[key].lower():
                        result.append(datacls)
        return BaseList(result)

    def __str__(self) -> str:
        return super().__str__()



import re
from typing import Dict, Any
from valorant_api.base_list import BaseList
from valorant_api.exceptions import AttributeExistsError


def snakeit(string: str):
    return re.sub(r'(?<!^)(?=[A-Z])', '_', string).lower()


class Mapper(object): # why
    raw_data: Dict[str, Any]

    def __init__(self):
        pass

    def map(self, childclasses=None, childdata=None):
        if childclasses is None:
            childclasses = {}
        raw_data = childdata if childdata is not None else self.raw_data
        for key, value in raw_data.items():
            key_snake = snakeit(key)
            if hasattr(self, key_snake):
                if getattr(self, key_snake) is not None:
                    raise AttributeExistsError(f"{self} already has '{key_snake}' attribute.")
            if isinstance(value, dict):
                child = childclasses[key]()
                child.map(childdata=value, childclasses=childclasses)
                setattr(self, key_snake, child)
            elif isinstance(value, list):
                lst = []
                for x in value:
                    child = childclasses[key]()
                    child.map(childdata=x, childclasses=childclasses)
                    lst.append(child)
                    setattr(self, key_snake, BaseList(lst))
            else:        
                setattr(self, key_snake, value)



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Level:
    uuid: str
    charm_level: int
    display_name: Optional[str]
    display_icon: Optional[str]
    asset_path: str

    def __init__(self, data: dict) -> None:
        self.uuid = data.get("uuid")
        self.charm_level = data.get("charmLevel")
        self.display_name = data.get("displayName")
        self.display_icon = data.get("displayIcon")
        self.asset_path = data.get("assetPath")


@dataclass
class Buddy:
    uuid: str
    display_name: str
    is_hidden_if_not_owned: bool
    theme_uuid: None
    display_icon: str
    asset_path: str
    levels: List[Level]
    raw_data: dict

    def __init__(self, data: dict):
        self.uuid = data.get("uuid")
        self.display_name = data.get("displayName")
        self.is_hidden_if_not_owned = data.get("isHiddenIfNotOwned")
        self.theme_uuid = data.get("themeUuid")
        self.display_icon = data.get("displayIcon")
        self.asset_path = data.get("assetPath")
        self.levels = [Level(x) for x in data.get("levels")]
        self.raw_data = data



from enum import Enum
from dataclasses import dataclass
from typing import List
from .utils import Mapper

# class TypeEnum(Enum):
#     CURRENCY = "Currency"
#     EQUIPPABLE_CHARM_LEVEL = "EquippableCharmLevel"
#     EQUIPPABLE_SKIN_LEVEL = "EquippableSkinLevel"
#     PLAYER_CARD = "PlayerCard"
#     SPRAY = "Spray"
#     TITLE = "Title"


@dataclass
class Reward(Mapper):
    type: str
    uuid: str
    amount: int
    is_highlighted: bool

    def __init__(self):
        super().__init__()

@dataclass
class Level(Mapper):
    reward: Reward
    xp: int
    vp_cost: int
    is_purchasable_with_vp: bool

    def __init__(self):
        super().__init__()

@dataclass
class Chapter(Mapper):
    is_epilogue: bool
    levels: List[Level]
    free_rewards: List[Reward]

    def __init__(self):
        super().__init__()

@dataclass
class Content(Mapper):
    relation_uuid: str
    relation_type: str
    chapters: List[Chapter]
    premium_reward_schedule_uuid: str
    premium_vp_cost: int

    def __init__(self):
        super().__init__()

@dataclass
class Contract(Mapper):
    uuid: str
    display_name: str
    display_icon: str
    ship_it: bool
    free_reward_schedule_uuid: str
    content: Content
    asset_path: str

    def __init__(self, data: dict) -> None:
        self.raw_data = data
        self.map(childclasses={"content": Content, "chapters": Chapter, "levels": Level, "reward": Reward, "freeRewards": Reward})




from dataclasses import dataclass
from datetime import datetime
from valorant_api.utils import Mapper


@dataclass
class Event(Mapper):
    uuid: str
    display_name: str
    short_display_name: str
    start_time: datetime
    end_time: datetime
    asset_path: str

    def __init__(self, data: dict) -> None:
        self.raw_data = data
        self.map(childclasses={})
